"""
Module used for dataclass logic to store related information in each Holder.
"""

from dataclasses import dataclass


@dataclass
class HOLDER_ENV_INFO:
    """
    A class to represent environment information for the holder.

    Attributes:
        level (int): The security or access level of the holder.
        uid_hash (str): The hashed user identifier associated with the holder.
        sig_hash (str): The hashed signature associated with the holder.
        encrypt_alg (dict): A dictionary representing the encryption algorithm(s) used.
        revoc_prob (float): The probability of the certificate being revoked.
        cert_valid_dur (str): The duration for which the certificate is valid.
        cache_dur (str): The duration for which data is cached.
        cooldown_dur (str): The cooldown duration between certain operations.
        timeout_dur (str): The duration after which an operation times out.
    """

    level: int
    uid_hash: str
    sig_hash: str
    encrypt_alg: dict
    revoc_prob: float
    cert_valid_dur: str
    cache_dur: str
    cooldown_dur: str
    timeout_dur: str


@dataclass
class HOLDER_TYPE_INFO:
    """
    A class to represent information about the type of holder's environment.

    Attributes:
        hardware_type (str): The type of hardware (e.g., endpoint, appliance).
        hardware_subtype (str): The subtype of the hardware (e.g., laptop, server).
        hardware_brand (str): The brand of the hardware (e.g., Dell, Cisco).
        os_category (str): The category of the operating system (e.g., Windows, Unix).
        os_subcategory (str): The subcategory of the operating system (e.g., Linux, Solaris).
        os_dist (str): The distribution of the operating system (e.g., Ubuntu, Windows 10).
        os_subdist (str): The subdistribution of the operating system (e.g., Alpine, Standard).
        account_type (str): The type of account (e.g., admin, user).
        account_subtype (str): The subtype of the account (e.g., network admin, personal).
        ca_status (str): The certificate authority status (e.g., root auth, revoked).

    Methods:
        long_name() -> str:
            Returns a detailed, formatted string combining various environment attributes for identification.

        short_name() -> str:
            Returns a shortened, formatted string for a more concise identification.
    """

    hardware_type: str
    hardware_subtype: str
    hardware_brand: str
    os_category: str
    os_subcategory: str
    os_dist: str
    os_subdist: str
    account_type: str
    account_subtype: str
    ca_status: str

    @property
    def long_name(self) -> str:
        """
        Returns the long version of the holder's type information as a property.

        Returns:
            str: The combined string of everything.
        """

        return f'{self.hardware_type}_{self.hardware_subtype}_{self.hardware_brand}.' \
               f'{self.os_category}_{self.os_subcategory}_{self.os_dist}_{self.os_subdist}.' \
               f'{self.account_type}_{self.account_subtype}.{self.ca_status}'

    @property
    def short_name(self) -> str:
        """
        Returns the short version of the holder's type information as a property.

        Returns:
            str: The combined string of the key elements of type information.
        """

        return f'{self.hardware_brand}.{self.os_subdist}.{self.account_subtype}.{self.ca_status}'


@dataclass
class HOLDER_INFO:
    """
    A class to represent detailed information about a holder, typically used in certificate generation.

    Attributes:
        common_name (str): The common name of the holder (e.g., individual or organization name).
        country (str): The country of the holder.
        state (str): The state or region of the holder.
        local (str): The local or city of the holder.
        org (str): The organization of the holder.
        org_unit (str): The organizational unit or department within the organization.
        email (str): The email address of the holder.
        url (str): The URL associated with the holder.

    Methods:
        hash_content() -> str:
            Returns a concatenated string of the holder's attributes
            that can be used for hashing or generating a unique identifier.
    """

    common_name: str
    country: str
    state: str
    local: str
    org: str
    org_unit: str
    email: str
    url: str

    @property
    def hash_content(self) -> str:
        """
        Returns all the content of the holder information as a hashable string.
        """

        return f'{self.common_name}' \
               f'{self.country}{self.state}{self.local}' \
               f'{self.org}{self.org_unit}{self.email}{self.url}'



"""
This file contains enums used in the program along with functions to retrieve information.
"""

import random
import copy
from enum import Enum
from typing import Union, List, Tuple


class SUPPORTED_HASH_ALGS(Enum):
    """
    This class contains all supported hash algorithms.
    It is used in both the autoconfiguration and manual configuration to set hashing algorithms.
    """
    SHA224 = 'sha224'
    SHA256 = 'sha256'
    SHA384 = 'sha384'
    SHA512 = 'sha512'
    SHA3_224 = 'sha3_224'
    SHA3_256 = 'sha3_256'
    SHA3_384 = 'sha3_384'
    SHA3_512 = 'sha3_512'
    BLAKE2B = 'blake2b'
    BLAKE2S = 'blake2s'


class SUPPORTED_ENCRYPT_ALGS(Enum):
    """
    This class contains all supported encryption algorithms.
    It is used in both the autoconfiguration and manual configuration to set encryption algorithms.
    """
    RSA = 'rsa'
    ECC = 'ecc'


class SUPPORTED_ECC_CURVES(Enum):
    """
    This class contains all supported curves for eliptic curve cryptography.
    It is used in both the autoconfiguration and manual configuration to set the curve for ECC.
    """
    SECP256R1 = 'secp256r1'
    SECP384R1 = 'secp384r1'
    SECP521R1 = 'secp521r1'
    SECP224R1 = 'secp224r1'
    SECP192R1 = 'secp192r1'
    SECP256K1 = 'secp256k1'


class COMMON_CA(Enum):
    """
    This class contains all supported certificate authority types by default.
    It is used in the manual configuration file to specify authority type.
    """
    NOT_AUTH = 'not_auth'
    INTER_AUTH = 'inter_auth'
    ROOT_AUTH = 'root_auth'


class COMMON_USERS(Enum):
    """
    This class contains all supported user types by default.
    It is used in the manual configuration file to specify user types.
    """
    GUEST = 'guest'
    PERSONAL = 'personal'
    ENTERPRISE = 'enterprise'


class COMMON_ADMINS(Enum):
    """
    This class contains all supported admin types by default.
    It is used in the manual configuration file to specify admin types.
    """
    DOMAIN_ADMIN = 'domain_admin'
    SCHEMA_ADMIN = 'schema_admin'
    SERVER_ADMIN = 'server_admin'
    NETWORK_ADMIN = 'network_admin'
    CLOUD_ADMIN = 'cloud_admin'
    DATABASE_ADMIN = 'database_admin'
    AUDITOR = 'auditor'


class COMMON_ACCOUNTS(Enum):
    """
    This class contains all supported account types by default.
    It is used in the manual configuration file to specify account types.
    """
    USER = 'user'
    ADMIN = 'admin'
    SYSTEM = 'system'


class COMMON_WINDOWS(Enum):
    """
    This class contains all supported Windows versions by default.
    It is used in the manual configuration file to specify Windows versions.
    """
    WINDOWS_2000 = ('professional', 'server', 'advanced_server', 'datacenter_server')
    WINDOWS_XP = ('home', 'professional')
    WINDOWS_VISTA = ('starter', 'home_basic', 'home_premium', 'business', 'enterprise', 'ultimate')
    WINDOWS_7 = ('starter', 'home_basic', 'home_premium', 'business', 'enterprise', 'ultimate')
    WINDOWS_8 = ('home', 'pro', 'enterprise')
    WINDOWS_10 = ('home', 'pro', 'educational', 'enterprise')
    WINDOWS_11 = ('home', 'pro', 'educational', 'enterprise')


class COMMON_WINDOWS_SERVER(Enum):
    """
    This class contains all supported Windows Server versions by default.
    It is used in the manual configuration file to specify Windows Server versions.
    """
    WINDOWS_SERVER_2003 = ('web', 'standard', 'enterprise', 'datacenter')
    WINDOWS_SERVER_2008 = ('web', 'standard', 'enterprise', 'datacenter', 'itanium', 'foundation', 'hpc')
    WINDOWS_SERVER_2012 = ('foundation', 'essentials', 'standard', 'datacenter')
    WINDOWS_SERVER_2016 = ('standard', 'datacenter')
    WINDOWS_SERVER_2019 = ('standard', 'datacenter')
    WINDOWS_SERVER_2022 = ('standard', 'datacenter', 'datacenter_azure')


class COMMON_MICROSOFT(Enum):
    """
    This class contains all supported Microsoft products by default.
    It is used in the manual configuration file to specify Microsoft products.
    """
    WINDOWS = 'windows'
    WINDOWS_SERVER = 'windows_server'


class COMMON_LINUX(Enum):
    """
    This class contains all supported Linux distributions by default.
    It is used in the manual configuration file to specify Linux distributions.
    """
    DEBIAN = ('debian', 'linux_mint', 'kali_linux', 'raspberry_pi', 'mx_linux', 'debian')
    RED_HAT = ('red_hat', 'fedora', 'cent_os')
    ARCH_LINUX = 'arch_linux'
    GENTOO = 'gentoo'
    SUSE = ('suse_linux_enterprise', 'open_suse')
    ALPINE = 'alpine'
    NIX_OS = 'nix_os'
    QUBES_OS = 'qubes_os'
    UBUNTU_SERVER = 'ubuntu_server'


class COMMON_BSD(Enum):
    """
    This class contains all supported BSD distributions by default.
    It is used in the manual configuration file to specify BSD distributions.
    """
    FREE_BSD = 'free_bsd'
    OPEN_BSD = 'open_bsd'
    NET_BSD = 'net_bsd'


class COMMON_MAC_OS_X(Enum):
    """
    This class contains all supported Mac OS X versions by default.
    It is used in the manual configuration file to specify Mac OS X versions.
    """
    LEOPARD = 'leopard'
    SNOW_LEOPARD = 'snow_leopard'
    LION = 'lion'
    MOUNTAIN_LION = 'mountain_lion'
    MAVERICKS = 'mavericks'
    YOSEMITE = 'yosemite'
    EL_CAPITAN = 'el_capitan'
    SIERRA = 'sierra'
    HIGH_SIERRA = 'high_sierra'
    MOJAVE = 'mojave'
    CATALINA = 'catalina'
    BIG_SUR = 'big_sur'
    MONTEREY = 'monterey'
    VENTURA = 'ventura'
    SONOMA = 'sonoma'
    SEQUOIA = 'sequoia'


class COMMON_UNIX(Enum):
    """
    This class contains all supported Unix flavors by default.
    It is used in the manual configuration file to specify Unix flavors.
    """
    LINUX = 'linux'
    BSD = 'bsd'
    SOLARIS = 'solaris'
    MAC_OS_X = 'mac_os_x'


class COMMON_MOBILE(Enum):
    """
    This class contains all supported mobile operating system families by default.
    It is used in the manual configuration file to specify mobile operating system families.
    """
    IOS = 'ios'
    ANDROID = (
        'android_nougat', 'android_oreo', 'android_pie', 'android_10', 'android_11', 'android_12',
        'android_13', 'android_14', 'android_15', 'android_16'
    )


class COMMON_ROUTING(Enum):
    """
    This class contains all supported routing platforms by default.
    It is used in the manual configuration file to specify routing platforms.
    """
    ONIE = 'onie'
    ONL = 'onl'
    OPX = 'opx'
    DNOS = 'dnos'
    JUNOS = 'junos'
    FBOSS = 'fboss'
    SONIC = 'sonic'
    ARUBA_OS = 'aruba_os'
    CISCO_IOS = 'cisco_ios'
    NEXUS_NOS = 'nexus_nos'
    OPENWRT = 'openwrt'


class COMMON_OS(Enum):
    """
    This class contains all supported operating system types by default.
    It is used in the manual configuration file to specify operating system types.
    """
    MICROSOFT = 'microsoft'
    UNIX = 'unix'
    MOBILE = 'mobile'
    ROUTING = 'routing'


class COMMON_ENDPOINT(Enum):
    """
    This class contains all supported endpoint hardware manufacturers by default.
    It is used in the manual configuration file to specify endpoint hardware manufacturers.
    """
    DESKTOP = ('hewlett_packard', 'acer', 'dell', 'lenovo', 'toshiba', 'ibm', 'fujitsu', 'nec', 'apple')
    LAPTOP = ('samsung', 'razer', 'microsoft', 'msi', 'asus', 'acer', 'dell', 'lenovo', 'hewlett_packard', 'apple')
    PHONE = ('samsung', 'apple', 'huawei', 'sony', 'google', 'microsoft', 'toshiba', 'dell')
    SERVER = ('dell', 'hewlett_packard', 'supermicro', 'inspur', 'lenovo', 'huawei', 'ibm', 'fukitsu', 'cisco')
    IOT = (
        'advantech', 'raspberry_pi', 'arudino', 'nvidia', 'beagleboard',
        'udoo', 'onlogic', 'kontron', 'arbor', 'axiomtek'
    )


class COMMON_NETWORK(Enum):
    """
    This class contains all supported network hardware manufacturers by default.
    It is used in the manual configuration file to specify network hardware manufacturers.
    """
    ROUTER = ('cisco', 'peplink', 'advantech', 'netgear', 'tp_link')
    SWITCH = ('anchor', 'honeywell', 'philips', 'siemens', 'cisco', 'hpl')
    ACCESS_POINT = ('cisco', 'fortinet', 'netgear', 'zyxel', 'tp_link', 'engenius')


class COMMON_APPLIANCE(Enum):
    """
    This class contains all supported appliance hardware manufacturers by default.
    It is used in the manual configuration file to specify appliance hardware manufacturers.
    """
    FIREWALL = ('bitdefender', 'cisco', 'fortinet', 'palo_alto', 'netgate', 'watchguard', 'sonicwall')
    UTM = ('sonicwall', 'fortigate', 'barracuda', 'juniper', 'trellix', 'palo_alto')


class COMMON_PERIPHERALS(Enum):
    """
    This class contains all supported peripheral hardware manufacturers by default.
    It is used in the manual configuration file to specify peripheral hardware manufacturers.
    """
    USB_KEY = ('samsung', 'sandisk', 'corsiar', 'kingston', 'pny')
    SMART_CARD = ('thales', 'nxp', 'cardlogix', 'infineon')
    EXTERNAL_STORAGE = ('seagate', 'western_digital', 'sandisk', 'transcend', 'lacie')


class COMMON_HARDWARE(Enum):
    """
    This class contains all supported hardware types by default.
    It is used in the manual configuration file to specify hardware types.
    """
    ENDPOINT = 'endpoint'
    NETWORK = 'network'
    APPLIANCE = 'appliance'
    PERIPHERAL = 'peripheral'


def has_value(enum_class, value: str) -> bool:
    """
    Check if a given value exists in the specified enum class.

    Args:
        enum_class: The enum class to check.
        value (str): The value to check.

    Returns:
        bool: True if the value exists in the enum class, False otherwise.
    """

    for item in enum_class:
        if isinstance(item.value, tuple):
            for v in item.value:
                if value == v.lower().replace(' ', '_').replace('-', '_'):
                    return True
        else:
            if value == item.value.lower().replace(' ', '_').replace('-', '_'):
                return True

    return False


def get_all_items(enum_class, verbose: bool = False) -> Union[dict, list]:
    """
    Return the versions of an enum class.
    If verbose is True, return the versions as a dictionary with names and values.
    If verbose is False, return the versions as a list with names only.

    Args:
        enum_class: The enum class to get the versions from.
        verbose (bool): If True, return the versions as a dictionary with names and values.
                        If False, return the versions as a list with names only.

    Returns:
        Union[dict, list]: The versions of the enum class.
    """

    if verbose:
        return {item.name: item.value for item in enum_class}
    else:
        return [item.name for item in enum_class]


def pass_rule_check(cur_settings: list) -> bool:
    """
    An arbitrary set of rule checks to make sure the random generation is at least realistic.

    Args:
        cur_settings (list): The current settings.

    Returns:
        bool: True if the current value passes the rule check, False otherwise.
    """

    h_type: str = cur_settings[0][0]
    h_subtype: str = cur_settings[0][1]
    h_brand: str = cur_settings[0][2]
    os_type: str = cur_settings[1][0]
    os_subtype: str = cur_settings[1][1]
    os_dist: str = cur_settings[1][2]
    a_type: str = cur_settings[2][0]
    ca_type: str = cur_settings[3][0]

    # Misplace checks
    # The hardware brand cannot be the hardware subtype
    brand_is_subtype = h_brand in [
        'desktop', 'laptop', 'server', 'phone', 'iot', 'switch', 'router', 'access_point', 'firewall', 'utm',
        'usb_key', 'smart_card', 'external_storage'
    ]
    if brand_is_subtype:
        return False

    # Type to subtype check
    bad_endpoint = h_type == 'endpoint' and h_subtype not in ['desktop', 'laptop', 'server', 'phone', 'iot', '']
    bad_network = h_type == 'network' and h_subtype not in ['switch', 'router', 'access_point', '']
    bad_appliance = h_type == 'appliance' and h_subtype not in ['firewall', 'utm', '']
    bad_peripheral = h_type == 'peripheral' and h_subtype not in ['usb_key', 'smart_card', 'external_storage', '']

    if bad_endpoint or bad_network or bad_appliance or bad_peripheral:
        return False

    # Hardware-to-Software rules
    # If a device type is a networking device, it can only use routing OSes and Unix OSes that are not Mac OS X
    network_is_routing = h_type == 'network' and os_type in ['routing', '']

    network_is_unix = h_type == 'network' and os_type in ['unix', '']
    unix_network_not_mac = network_is_unix and os_subtype != 'mac_os_x'

    if h_type == 'network' and not (network_is_routing or unix_network_not_mac):
        return False

    # If a device endpoint type is an IoT device, it can only use Unix OSes that are not Mac OS X
    iot_is_unix = h_subtype == 'iot' and os_type in ['unix', '']
    unix_iot_not_mac = iot_is_unix and os_subtype != 'mac_os_x'
    if h_subtype == 'iot' and not unix_iot_not_mac:
        return False

    # If a device endpoint type is a phone, it can only use mobile operating systems
    phone_is_mobile = h_subtype == 'phone' and os_type in ['mobile', '']
    if h_subtype == 'phone' and not phone_is_mobile:
        return False

    # If a device endpoint type is a server, it can only use Windows Server OSes and Unix OSes that are not Mac OS X
    server_is_microsoft = h_subtype == 'server' and os_type in ['microsoft', '']
    ms_server_is_ws = server_is_microsoft and os_subtype in ['windows_server', '']

    server_is_unix = h_subtype == 'server' and os_type in ['unix', '']
    unix_server_not_mac = server_is_unix and os_subtype != 'mac_os_x'

    if h_subtype == 'server' and not (ms_server_is_ws or unix_server_not_mac):
        return False

    # If a device endpoint type is a laptop or desktop it can't use a mobile OS
    laptop_or_desktop = h_subtype == 'laptop' or h_subtype == 'desktop'
    lap_desk_not_mobile = laptop_or_desktop and os_type != 'mobile'
    if laptop_or_desktop and not lap_desk_not_mobile:
        return False

    # Hardware-to-Account rules
    # A phone cannot have an admin account
    phone_is_not_admin = h_subtype == 'phone' and a_type != 'admin'
    if h_subtype == 'phone' and not phone_is_not_admin:
        return False

    # A server cannot have a user account
    server_is_not_user = h_subtype == 'server' and a_type != 'user'
    if h_subtype == 'server' and not server_is_not_user:
        return False

    # Networking devices cannot use user accounts
    net_is_not_user = h_type == 'network' and a_type != 'user'
    if h_type == 'network' and not net_is_not_user:
        return False

    # Appliances and Peripherals have to be a system account
    app_or_peri = h_type == 'appliance' or h_type == 'peripheral'
    app_peri_is_system = app_or_peri and a_type in ['system', '']
    if app_or_peri and not app_peri_is_system:
        return False

    # Software-to-Hardware rules
    # Mobile OSs can only be endpoints
    mobile_uses_endpoint = os_type == 'mobile' and h_type in ['endpoint', '']
    if os_type == 'mobile' and not mobile_uses_endpoint:
        return False

    # Mobile OSs can only use phones
    mobile_uses_phone = os_type == 'mobile' and h_subtype in ['phone', '']
    if os_type == 'mobile' and not mobile_uses_phone:
        return False

    # Mac OS X can only be endpoints
    mac_uses_endpoint = os_subtype == 'mac_os_x' and h_type in ['endpoint', '']
    if os_subtype == 'mac_os_x' and not mac_uses_endpoint:
        return False

    # Mac OS X can only be on desktops and laptops
    mac_uses_pc = os_subtype == 'mac_os_x' and h_subtype in ['desktop', 'laptop', '']
    if os_subtype == 'mac_os_x' and not mac_uses_pc:
        return False

    # Windows can only be on endpoints
    windows_uses_endpoint = os_subtype == 'windows' and h_type in ['endpoint', '']
    if os_subtype == 'windows' and not windows_uses_endpoint:
        return False

    # Windows can only be on desktops and laptops
    windows_uses_pc = os_subtype == 'windows' and h_subtype in ['desktop', 'laptop', '']
    if os_subtype == 'windows' and not windows_uses_pc:
        return False

    # Software-to-Account rules
    # If Windows Server OS it has to be a system or admin account
    ws_is_system_or_admin = os_subtype == 'windows_server' and a_type in ['system', 'admin', '']
    if os_subtype == 'windows_server' and not ws_is_system_or_admin:
        return False

    # If Ubuntu Server OS it has to be a system or admin account
    us_is_system_or_admin = os_dist == 'ubuntu_server' and a_type in ['system', 'admin', '']
    if os_dist == 'ubuntu_server' and not us_is_system_or_admin:
        return False

    # Routing OSes cannot have user accounts
    routing_is_not_user = os_type == 'routing' and a_type != 'user'
    if os_type == 'routing' and not routing_is_not_user:
        return False

    # Mobile OSes cannot have admin accounts
    mobile_is_not_admin = os_type == 'mobile' and a_type != 'admin'
    if os_type == 'mobile' and not mobile_is_not_admin:
        return False

    # Account-to-Hardware rules
    # User accounts can only use desktops, laptops, and phones
    user_is_user_device = a_type == 'user' and h_subtype in ['desktop', 'laptop', 'phone', '']
    if a_type == 'user' and not user_is_user_device:
        return False

    # Admin accounts cannot use peripherals
    admin_is_not_peri = a_type == 'admin' and h_type != 'peripheral'
    if a_type == 'admin' and not admin_is_not_peri:
        return False

    # Account-to-Software rules
    # User accounts cannot exist on routing OSes or Ubuntu Server
    user_is_not_routing = a_type == 'user' and os_type != 'routing'
    user_is_not_us = a_type == 'user' and os_dist != 'ubuntu_server'
    if a_type == 'user' and not (user_is_not_routing and user_is_not_us):
        return False

    # CA based rules
    # If Not a Server it cannot be a certificate authority
    is_ca = ca_type in ['inter_auth', 'root_auth']
    ca_is_endpoint = is_ca and h_type in ['endpoint', '']
    ca_is_server = is_ca and h_subtype in ['server', '']
    if is_ca and not (ca_is_endpoint and ca_is_server):
        return False

    return True


def check_exceptions(cur_settings: list) -> list:
    """
    Pre-emptively fills out the settings if it runs into an exception rule.

    Args:
        cur_settings (list): The current settings.

    Returns:
        list: The pre-filled list if needed.
    """

    h_type: str = cur_settings[0][0]
    os_type: str = cur_settings[1][0]
    os_subtype: str = cur_settings[1][1]
    os_dist: str = cur_settings[1][2]
    a_type: str = cur_settings[2][0]

    # Hardware exceptions
    # If an appliance or peripheral, the all OS types should be the specific brand followed by "_os"
    if h_type in ['appliance', 'peripheral']:
        cur_settings[1][0] = cur_settings[0][2] + '_os'
        cur_settings[1][1] = cur_settings[0][2] + '_os'
        cur_settings[1][2] = cur_settings[0][2] + '_os'
        cur_settings[1][3] = cur_settings[0][2] + '_os'

    # Software exceptions
    # If not windows, windows-server, suse, red-hat, or debian dist, then OS subdist should be the OS dist.
    os_is_microsoft = os_type == 'microsoft'
    linux_os_has_subdist = os_dist in ['suse', 'red_hat', 'debian']

    if not (os_is_microsoft or linux_os_has_subdist):
        cur_settings[1][3] = os_dist

    # If a routing os, mobile-ios, or solaris, then all OS types should be the routing os or ios
    if os_type == 'routing' or os_subtype == 'ios' or os_subtype == 'solaris':
        cur_settings[1][2] = os_subtype
        cur_settings[1][3] = os_subtype

    # Account exceptions
    # If system, then the account subtype should be system
    if a_type == 'system':
        cur_settings[2][1] = a_type

    return cur_settings


def check_enum_value(outer_index: int, inner_index: int, cur_settings: list,
                     key: str, value: Union[tuple, str], parent: bool = False) -> tuple:
    """
    Decision Tree to check the values of the chosen enum given the enum key and the value(s).

    Args:
        outer_index (int): The index of the outer list.
        inner_index (int): The index of the inner list.
        cur_settings (list): The current settings list.
        key (str): The chosen key from the returned enum dictionary.
        value (Union[tuple, str]): The values of the chosen key.
        parent (bool): Whether a parent was used to get here.

    Returns:
        tuple: A tuple of-
            If the tested value was validated.
            A list of the current settings regardless of modification.
    """

    # Set variables
    temp_settings: list = copy.deepcopy(cur_settings)
    new_value_unvalidated: bool = True

    # Set the inner index for the check
    if parent:
        local_inner_index: int = inner_index - 1
    else:
        local_inner_index: int = inner_index

    # Check if the value is a tuple and work accordingly
    if isinstance(value, tuple):
        # Create a list to go through
        end_options: list = list(value)
        using_end_options: bool = True

        # Try options in list
        while end_options and using_end_options:
            # Set options
            temp_settings[outer_index][local_inner_index] = key.lower()

            end_option_index: int = random.randint(0, len(end_options) - 1)
            temp_settings[outer_index][local_inner_index + 1] = end_options[end_option_index]

            # Check if temporary settings passes the rule check
            if pass_rule_check(temp_settings):
                cur_settings = temp_settings
                using_end_options = False
                new_value_unvalidated = False

            # Remove used option
            del end_options[end_option_index]
    else:
        temp_settings[outer_index][local_inner_index] = value

        # Check if temporary settings passes the rule check
        if pass_rule_check(temp_settings):
            cur_settings = temp_settings
            new_value_unvalidated = False

    return new_value_unvalidated, cur_settings


def update_settings(outer_index: int, inner_index: int, cur_settings: list, locked_settings: tuple) -> list:
    """
    Updates the settings list based on the current information being looked at.

    Args:
        outer_index (int): The index of the outer list.
        inner_index (int): The index of the inner list.
        cur_settings (list): The current settings list.
        locked_settings (tuple): The settings that MUST stay.

    Returns:
        list: The updated settings list.
    """

    enum_switch_known: dict = {
        'endpoint': COMMON_ENDPOINT,
        'network': COMMON_NETWORK,
        'appliance': COMMON_APPLIANCE,
        'peripheral': COMMON_PERIPHERALS,
        'routing': COMMON_ROUTING,
        'mobile': COMMON_MOBILE,
        'unix': COMMON_UNIX,
        'microsoft': COMMON_MICROSOFT,
        'windows': COMMON_WINDOWS,
        'windows_server': COMMON_WINDOWS_SERVER,
        'bsd': COMMON_BSD,
        'linux': COMMON_LINUX,
        'mac_os_x': COMMON_MAC_OS_X,
        'admin': COMMON_ADMINS,
        'user': COMMON_USERS
    }

    enum_switch_unknown: dict = {
        0: {
            0: [COMMON_HARDWARE],
            1: [COMMON_PERIPHERALS, COMMON_APPLIANCE, COMMON_NETWORK, COMMON_ENDPOINT]
        },
        1: {
            0: [COMMON_OS],
            1: [COMMON_ROUTING, COMMON_MOBILE, COMMON_UNIX, COMMON_MICROSOFT],
            2: [COMMON_WINDOWS_SERVER, COMMON_WINDOWS, COMMON_BSD, COMMON_LINUX, COMMON_MAC_OS_X]
        },
        2: {
            0: [COMMON_ACCOUNTS],
            1: [COMMON_ADMINS, COMMON_USERS]
        },
        3: {
            0: [COMMON_CA]
        }
    }

    parent_value: str = ''
    last_value: str = ''
    new_value_unvalidated: bool = True

    # Choose the next enum if the last value is not empty
    if inner_index > 0 and cur_settings[outer_index][inner_index-1] != '':  # Go down a certain path
        last_value = cur_settings[outer_index][inner_index-1]

        # If setting the last value was successful, set the enum to use.
        if last_value in enum_switch_known.keys():
            enum_choice = enum_switch_known[last_value]
        else:
            # If not, use the parent to get the enum data
            parent_value: str = cur_settings[outer_index][inner_index-2]

            # If the parent's found, set the enum_choice
            if parent_value in enum_switch_known.keys():
                enum_choice = enum_switch_known[parent_value]
            else:
                # If the parent is not found, fill with whatever value was detected and call it a day
                for i in range(inner_index, len(cur_settings[outer_index])):
                    if cur_settings[outer_index][inner_index] == '':
                        cur_settings[outer_index][inner_index] = last_value

                # Address Hard-fill cases
                cur_settings = check_exceptions(cur_settings)

                # Re-enter locks
                for lock_i in range(len(locked_settings)):
                    for lock_j in range(len(locked_settings[lock_i])):
                        if locked_settings[lock_i][lock_j] != '':
                            cur_settings[lock_i][lock_j] = locked_settings[lock_i][lock_j]

                return cur_settings
    else:  # Random selection as final solution
        enum_type_loc: list = enum_switch_unknown[outer_index][inner_index]
        enum_index: int = random.randint(0, len(enum_type_loc) - 1)
        enum_choice = enum_type_loc[enum_index]

    # Get the enum data
    enum_data: dict = get_all_items(enum_choice, verbose=True)

    # If parent value set, go straight through the parent for the attempt
    if parent_value != '':
        # Set the child value and test it
        if last_value.upper() in enum_data.keys():
            child_value = enum_data[last_value.upper()]
            new_value_unvalidated, cur_settings = check_enum_value(
                outer_index, inner_index, cur_settings, last_value, child_value, parent=True
            )

        else:  # If the parent is not found, fill with whatever value was detected and call it a day
            for i in range(inner_index, len(cur_settings[outer_index])):
                if cur_settings[outer_index][inner_index] == '':
                    cur_settings[outer_index][inner_index] = last_value
    else:
        # Get a random value from the enum data the abides by unknown enum switch and test.
        # Keep doing that until a value passes the rule set.
        while enum_data and new_value_unvalidated:
            # Get a random value
            random_key: str = random.choice(list(enum_data.keys()))
            random_value: str = enum_data[random_key]

            # Test the value
            new_value_unvalidated, cur_settings = check_enum_value(
                outer_index, inner_index, cur_settings, random_key, random_value
            )

            # Remove key once used
            del enum_data[random_key]

    # Address Hard-fill cases
    cur_settings = check_exceptions(cur_settings)

    # Re-enter locks
    for lock_i in range(len(locked_settings)):
        for lock_j in range(len(locked_settings[lock_i])):
            if locked_settings[lock_i][lock_j] != '':
                cur_settings[lock_i][lock_j] = locked_settings[lock_i][lock_j]

    return cur_settings


def auto_fill_types(cur_settings: list) -> Union[list, None]:
    """
    Autofill the types of the enum class.

    Args:
        cur_settings (list): The current settings list.

    Returns:
        tuple: Exception flag and randomly generated type properties for holder.
    """

    # Set locks
    locked_settings: tuple = tuple(copy.deepcopy(cur_settings))

    # Try the system
    try:
        for outer_index in range(len(cur_settings)):
            for inner_index in range(len(cur_settings[outer_index])):
                # Get the enum to try to change
                cur_info: str = cur_settings[outer_index][inner_index]
                if cur_info == '':
                    cur_settings = update_settings(outer_index, inner_index, cur_settings, locked_settings)
    except Exception as e:
        print(e)
        print(cur_settings)
        return None

    return cur_settings



"""
Module containing all utilities relevant for interacting with a command line interface.
"""

# Relative pathing from project root
import sys
from typing import Union
from os.path import abspath, dirname, join

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

from PKIPractice.Utilities.IngestUtils import *
from PKIPractice.Simulation.Network import PKINetwork


def get_default_auto() -> dict:
    """
    Retrieve the default autoconfiguration

    Returns:
        dict: The default autoconfiguration
    """
    auto_config: dict = {
        "level_count": 4,
        "count_by_level": [1, 2, 4, 8],
        "uid_hash": "sha256",
        "sig_hash": "sha256",
        "encrypt_alg": {
            "alg": "rsa",
            "params": {
                "pub_exp": 65537,
                "key_size": 2048
            }
        },
        "revoc_probs": [0.0, 0.0001, 0.001, 0.01],
        "cert_valid_durs": ["none", "00:15:00", "00:10:00", "00:05:00"],
        "cache_durs": ["none", "11:00", "06:00", "01:00"],
        "cooldown_durs": ["none", "5", "5", "5"],
        "timeout_durs": ["none", "20", "20", "20"],
        "log_save_filepath": "output/saved_network_logs_default.csv"
    }

    return auto_config


def get_default_manual() -> dict:
    """
    Retrieve the default manual configuration

    Returns:
        dict: The default manual configuration
    """
    manual_config: dict = {
        "default_root_ca": {
            "location": {
                "level": 1
            },
            "env_overrides": {
                "uid_hash": "sha3_512",
                "sig_hash": "sha3_512",
                "encrypt_alg": {
                    "alg": "ecc",
                    "params": {
                        "curve": "secp256r1"
                    }
                },
                "revoc_prob": 0.0,
                "cert_valid_dur": "none",
                "cache_dur": "none",
                "cooldown_dur": "none",
                "timeout_dur": "none"
            },
            "holder_type_info": {
                "hardware_type": "endpoint",
                "hardware_subtype": "server",
                "hardware_brand": "dell",
                "os_category": "microsoft",
                "os_subcategory": "windows_server",
                "os_dist": "windows_server_2019",
                "os_subdist": "standard",
                "account_type": "admin",
                "account_subtype": "domain_admin",
                "ca_status": "root_auth"
            },
            "holder_info": {
                "common_name": "Root Enterprises Root CA",
                "country": "US",
                "state": "CA",
                "locality": "San Francisco",
                "org": "Root Enterprises",
                "org_unit": "Certificates",
                "email": "root_ca_team@root_enterprises.com",
                "url": "root_enterprises.com/root_ca"
            }
        },
        "second_lvl_ca_one": {
            "location": {
                "level": 2
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512",
                "encrypt_alg": {
                    "alg": "ecc",
                    "params": {
                        "curve": "secp256r1"
                    }
                }
            },
            "holder_type_info": {
                "os_category": "unix",
                "os_subcategory": "linux",
                "os_dist": "ubuntu_server"
            }
        },
        "second_lvl_ca_two": {
            "location": {
                "level": 2
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512",
                "encrypt_alg": {
                    "alg": "ecc",
                    "params": {
                        "curve": "secp256r1"
                    }
                }
            },
            "holder_type_info": {
                "os_category": "unix",
                "os_subcategory": "linux",
                "os_dist": "ubuntu_server"
            }
        },
        "third_lvl_ca_one": {
            "location": {
                "level": 3
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512"
            },
            "holder_info": {
                "common_name": "Cert Incorporated South America CA",
                "country": "PE",
                "state": "Lima",
                "locality": "Ventanilla",
                "org": "Cert Incorporated",
                "org_unit": "South American Certificates",
                "email": "certs_sa@cert_incorporated.com",
                "url": "cert_incorporated.com/peru/intermediate_ca"
            }
        },
        "third_lvl_ca_two": {
            "location": {
                "level": 3
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512"
            },
            "holder_info": {
                "common_name": "CloudCert Inc West Africa CA",
                "country": "Nigeria",
                "state": "Oyo",
                "locality": "Ibadan",
                "org": "CloudCert Inc",
                "org_unit": "West African Certificates",
                "email": "certs_africa@cloudcert.com",
                "url": "cloudcert.com/nigeria/intermediate_ca"
            }
        },
        "third_lvl_ca_three": {
            "location": {
                "level": 3
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512"
            },
            "holder_info": {
                "common_name": "EuroPass International Norway CA",
                "country": "NO",
                "state": "Bergen",
                "locality": "Kokstad",
                "org": "EuroPass International",
                "org_unit": "Western European Certificates",
                "email": "certs_europe@europass.com",
                "url": "europass.com/norway_intermediate_ca"
            }
        },
        "third_lvl_ca_four": {
            "location": {
                "level": 3
            },
            "env_overrides": {
                "uid_hash": "sha512",
                "sig_hash": "sha512"
            },
            "holder_info": {
                "common_name": "Lone Star Networking Houston CA",
                "country": "US",
                "state": "Texas",
                "locality": "Houston",
                "org": "Lone Star Networking",
                "org_unit": "North American Certificates",
                "email": "lone_star_certs@lonestarnet.com",
                "url": "lonestarnet.com/us/houston/intermediate_ca"
            }
        },
        "fourth_level_one": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "network",
                "hardware_subtype": "access_point",
                "hardware_brand": "cisco",
                "os_category": "routing",
                "os_subcategory": "openwrt",
                "os_dist": "openwrt",
                "os_subdist": "openwrt",
                "account_type": "admin",
                "account_subtype": "network_admin"
            }
        },
        "fourth_level_two": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "endpoint",
                "hardware_subtype": "laptop",
                "hardware_brand": "asus",
                "os_category": "microsoft",
                "os_subcategory": "windows",
                "os_dist": "windows_10",
                "os_subdist": "home",
                "account_type": "user",
                "account_subtype": "personal"
            }
        },
        "fourth_level_three": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "peripheral",
                "hardware_subtype": "smart_card"
            }
        },
        "fourth_level_four": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "endpoint",
                "hardware_subtype": "phone",
                "account_type": "user"
            }
        },
        "fourth_level_five": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "appliance",
                "hardware_subtype": "utm",
                "hardware_brand": "barracuda"
            }
        },
        "fourth_level_six": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "endpoint",
                "hardware_subtype": "desktop",
                "os_category": "unix",
                "os_subcategory": "solaris",
                "account_subtype": "cloud_admin"
            }
        },
        "fourth_level_seven": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "hardware_type": "endpoint",
                "hardware_subtype": "iot",
                "hardware_brand": "arduino",
                "os_category": "unix",
                "os_subcategory": "linux",
                "os_dist": "alpine",
                "os_subdist": "alpine",
                "account_type": "user",
                "account_subtype": "guest"
            }
        },
        "fourth_level_eight": {
            "location": {
                "level": 4
            },
            "holder_type_info": {
                "os_subcategory": "mac_os_x"
            }
        }
    }

    return manual_config


def ingest_config(args: list, default: bool) -> Union[tuple, None]:
    """
    Starts the program using the command-line arguments.

    Args:
        args (list): A list of command-line arguments.
        default (bool, optional): If True, use the default configuration files. Defaults to False.
    """

    # Check if a yaml file is passed on an interpreter before Python 3.10
    if sys.version_info[1] < 10:
        assert all('.yaml' not in arg for arg in args), (
            'Invalid configuration filepath provided.\n'
            '\t   Yaml files do not have support for Python versions before 3.10.\n'
            '\t   Please use a different configuration format (JSON, XML, TOML).\n'
        )

    # Check if there is a proper argument for the auto generation
    if not default:
        assert 'auto' in args[1], (
            'Invalid configuration filepath provided.\n'
            '\t   Please provide a proper auto configuration file by '
            'passing the filepath of your file as an command-line argument.\n'
            '\t   Example: python Main.py Default_Configs/default_auto.yaml\n'
        )

    # Check if there is a proper argument for the manual settings or if it's just one argument
    only_auto_or_default: bool = len(args) == 2 or default
    if only_auto_or_default:
        manual_exists: bool = True
    else:
        manual_exists: bool = 'manual' in args[2]
    assert manual_exists is True, (
        'Invalid configuration filepath provided.\n'
        '\t   Please provide a proper manual configuration file by '
        'passing the filepath of your file as an command-line argument.\n'
        '\t   Example: python Main.py Default_Configs/default_auto.yaml Default_Configs/default_manual.yaml\n'
    )

    # Warn if there are more than the two arguments that have been checked
    if len(args) > 3:
        print('Warning: More than two command-line argument provided.\n'
              '\t Please provide a configuration file by '
              'passing the filepath of your file as an command-line argument.\n'
              '\t   Example: python Main.py Default_Configs/default_auto.yaml '
              'Default_Configs/default_manual.yaml\n')

    # Pass auto argument to ingestion utilities
    if default:
        env_auto_settings: Union[dict, None] = get_default_auto()
        assert validate_settings_auto(env_auto_settings) is True, (
            'Ingested autoconfiguration settings were not found to be valid.\n'
            '\t   Please ensure your configuration file is correctly created.\n'
            '\t   Use the default configuration file as a template.\n'
        )
    else:
        env_auto_settings: Union[dict, None] = parse_config_auto(args[1])

    # Pass manual argument to ingestion utilities
    if default:
        env_manual_settings: Union[dict, None] = get_default_manual()
        env_manual_settings = search_for_typecast_manual(env_manual_settings)
        assert env_manual_settings is not None, (
            'Ingested manual configuration settings were not able to be adjusted due to '
            'unparsable configuration params.\n'
            '\t   Please ensure your configuration file is correctly created.\n'
            '\t   Use the default configuration file as a template.\n'
        )
    else:
        if len(args) > 2:
            env_manual_settings: Union[dict, None] = parse_config_manual(args[2])
        else:
            env_manual_settings: Union[dict, None] = None

    # Check the return values for both
    assert env_auto_settings is not None, (
        'Unparseable autoconfiguration file provided.\n'
        '\t   Please ensure that your configuration file exists or are properly created.\n'
        '\t   Use the default configuration files provided in the Default_Configs folder as a guide.\n'
    )

    if len(args) > 2:
        assert env_manual_settings is not None, (
            'Unparseable manual configuration file provided.\n'
            '\t   Please ensure that your configuration file exists or are properly created.\n'
            '\t   Use the default configuration files provided in the Default_Configs folder as a guide.\n'
        )

    return env_auto_settings, env_manual_settings


def start_program() -> None:
    """
    Starts the program. Used by RunConfig.py and command line call to start program.
    """
    # Name flags
    help_flag = False
    test_flag = False
    default_flag = False

    # Start assertion region
    try:
        pki_network: Union[None, PKINetwork] = None

        # Check if there are more than one argument
        assert len(sys.argv) > 1, (
            'No configuration file provided.\n' 
            '\t   Please provide a configuration file by '
            'passing the filepath of your file as an command-line argument.\n'
            '\t   Example: python Main.py Default_Configs/default_auto.yaml\n'
        )

        # Check if there is a help flag
        if any(arg in sys.argv for arg in ('-h', '--help')):
            print(
                '   Help flag detected.\n'
                '   Welcome to PKI Practice!\n'
                '   This is not really meant for much, I just wanted to practice PKI architecture.\n'
                '   However, that does not mean that it should not be fun to play with.\n'
                '\n'
                '   In terms of command-line usage, you need to provide only to files.\n'
                '   The first is a configuration file for the auto generation of the environment.\n'
                '   The second is a configuration file for the manual configuration of the environment.\n'
                '   The second file is optional to run the program, but the first can be run without the second.\n'
                '\n'
                '   Structure: python Main.py [options] [<autoconfig filepath>] [<manualconfig filepath>]\n'
                '   Example without second: python Main.py Default_Configs/default_auto.yaml\n'
                '   Example with second: python Main.py Default_Configs/default_auto.yaml '
                'Default_Configs/default_manual.yaml\n'
                '\n'
                '   Options:\n'
                '   -h\n'
                '   --help\n'
                "   Prints the help information you see now.\n"
                '   --------\n'
                '   -d\n'
                '   --default\n'
                "   Tells the program to run it's default configuration\n"
                '   --------\n'
                '   -t\n'
                '   --test\n'
                "   Tells the program to run in test mode, where it only runs enough of the program to conduct "
                "assessments.\n"
                '\n\n'
                '   For more details, please check out https://laoluadewoye.github.io/PKI_Practice_Python/.\n'
            )
            help_flag = True
            help_index = next((i for i, arg in enumerate(sys.argv) if arg in ('-h', '--help')), None)
            sys.argv.pop(help_index)

        # Check if there is a test flag
        if any(arg in sys.argv for arg in ('-t', '--test')):
            test_flag = True
            test_index = next((i for i, arg in enumerate(sys.argv) if arg in ('-t', '--test')), None)
            sys.argv.pop(test_index)

        # Check if there is a default flag
        if any(arg in sys.argv for arg in ('-d', '--default')):
            print(
                '   Default flag detected.\n'
                '   Welcome to PKI Practice!\n'
                '   This is not really meant for much, I just wanted to practice PKI architecture.\n'
                '   However, that does not mean that it should not be fun to play with.\n'
                '\n'
                '   In terms of command-line usage, you need to provide only to files.\n'
                '   The first is a configuration file for the auto generation of the environment.\n'
                '   The second is a configuration file for the manual configuration of the environment.\n'
                '   The second file is optional to run the program, but the first can be run without the second.\n'
                '\n'
                "   For more details, please run this command with the help option [-h | --help] "
                "or check out https://laoluadewoye.github.io/PKI_Practice_Python/.\n"
                '\n'
                '   For now though, here is a default run of the program using the default yaml files.\n'
            )
            default_flag = True
            default_index = next((i for i, arg in enumerate(sys.argv) if arg in ('-d', '--default')), None)
            sys.argv.pop(default_index)

        # Start the program if nothing else is needed.
        if not help_flag:
            # Read the configuration files or default configurations
            env_auto_settings, env_manual_settings = ingest_config(sys.argv, default=default_flag)

            # Build the environment
            pki_network = PKINetwork('Sample_Net', env_auto_settings, env_manual_settings)

        # Go even further if not just testing the CLI options.
        if not test_flag and not help_flag:
            print(pki_network)
            pki_network.save_logs()

    # Ultimate error escape
    except AssertionError as e:
        print(f'\nException: {e}')



"""
Module used for defining the holder class and it's functionality.
"""

# Relative pathing from project root
import sys
from queue import PriorityQueue
from os.path import abspath, dirname, join
from typing import Union
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from .SimUtils import hash_info, get_random_country, get_random_division, create_private_key

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

from PKIPractice.Utilities.DataclassUtils import *
from PKIPractice.Utilities.EnumUtils import auto_fill_types


class PKIHolder:
    """
    The PKIHolder class represents a public key infrastructure (PKI) holder.

    Attributes:
        holder_name (str): The name of the PKI holder.
        env_info (HOLDER_ENV_INFO): Environment-related information for the PKI holder, such as
            encryption algorithm, UID hash, and various durations.
        holder_type_info (HOLDER_TYPE_INFO): Information about the type of holder, such as hardware
            type, operating system details, and certification authority status.
        holder_info (HOLDER_INFO): General information about the holder, including name, location,
            organization, and contact details.
        holder_info_hash (str): Hash of the holder general information for secure identification.
        holder_priv_key (Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey]): The holder's private key.
        holder_pub_key (Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]): The holder's public key.
        holder_cert: Placeholder for the holder's certificate.
        root_certs (dict): Dictionary storing root certificates.
        cached_certs (dict): Cached certificates for quick access.
        csr_message_port (PriorityQueue): Queue for CSR (Certificate Signing Request) messages.
        reg_message_port (PriorityQueue): Queue for registration messages.
        ocsp_message_port (PriorityQueue): Queue for OCSP (Online Certificate Status Protocol) messages.
        has_root_cert_cache (bool): Flag indicating whether root certificates are cached.
        has_self_cert (bool): Flag indicating whether the holder has a self-signed certificate.
        need_new_cert (bool): Flag indicating whether the holder requires a new certificate.
        cached_certs_empty (bool): Flag indicating whether cached certificates are empty.
        waiting_for_csr_response (bool): Indicates if the holder is awaiting a CSR response.
        waiting_to_send_csr (bool): Indicates if the holder is waiting to send a CSR.
        waiting_for_reg_response (bool): Indicates if the holder is awaiting a registration response.
        waiting_to_send_reg (bool): Indicates if the holder is waiting to send a registration request.
        waiting_for_ocsp_response (bool): Indicates if the holder is awaiting an OCSP response.
        waiting_to_send_oscp (bool): Indicates if the holder is waiting to send an OCSP request.
        network_hub: Connection to the network hub.
        is_ca (bool): Indicates if the holder is a certification authority (CA).
        lower_level_certs (dict): Stores certificates for lower-level entities.
        cert_revoc_list (dict): Stores the certificate revocation list.

    Methods:
        get_name() -> str:
            Returns the name of the name of the holder.
        set_hub_conn(hub) -> None:
            Sets the hub connection of the holder.
    """
    def __init__(self, holder_name: str, holder_config: dict, auto_config: dict):
        # Name of holder
        self.holder_name: str = holder_name

        level = holder_config['location']['level']

        # Environment information
        has_env_overrides = 'env_overrides' in holder_config.keys()
        if has_env_overrides and 'uid_hash' in holder_config['env_overrides'].keys():
            uid_hash = holder_config['env_overrides']['uid_hash']
        else:
            uid_hash = auto_config['uid_hash']

        if has_env_overrides and 'sig_hash' in holder_config['env_overrides'].keys():
            sig_hash = holder_config['env_overrides']['sig_hash']
        else:
            sig_hash = auto_config['sig_hash']

        if has_env_overrides and 'encrypt_alg' in holder_config['env_overrides'].keys():
            encrypt_alg = holder_config['env_overrides']['encrypt_alg']
        else:
            encrypt_alg = auto_config['encrypt_alg']

        if has_env_overrides and 'revoc_prob' in holder_config['env_overrides'].keys():
            revoc_prob = holder_config['env_overrides']['revoc_prob']
        else:
            revoc_prob = auto_config['revoc_probs'][level-1]

        if has_env_overrides and 'cert_valid_dur' in holder_config['env_overrides'].keys():
            cert_valid_dur = holder_config['env_overrides']['cert_valid_dur']
        else:
            cert_valid_dur = auto_config['cert_valid_durs'][level-1]

        if has_env_overrides and 'cache_dur' in holder_config['env_overrides'].keys():
            cache_dur = holder_config['env_overrides']['cache_dur']
        else:
            cache_dur = auto_config['cache_durs'][level-1]

        if has_env_overrides and 'cooldown_dur' in holder_config['env_overrides'].keys():
            cooldown_dur = holder_config['env_overrides']['cooldown_dur']
        else:
            cooldown_dur = auto_config['cooldown_durs'][level-1]

        if has_env_overrides and 'timeout_dur' in holder_config['env_overrides'].keys():
            timeout_dur = holder_config['env_overrides']['timeout_dur']
        else:
            timeout_dur = auto_config['timeout_durs'][level-1]

        self.env_info: HOLDER_ENV_INFO = HOLDER_ENV_INFO(
            level=level, uid_hash=uid_hash, sig_hash=sig_hash, encrypt_alg=encrypt_alg, revoc_prob=revoc_prob,
            cert_valid_dur=cert_valid_dur, cache_dur=cache_dur, cooldown_dur=cooldown_dur, timeout_dur=timeout_dur
        )

        # Type information
        type_fill = [['', '', ''], ['', '', '', ''], ['', ''], ['']]
        has_holder_type = 'holder_type_info' in holder_config.keys()

        if level == 1:
            type_fill[3][0] = 'root_auth'
        elif level < auto_config['level_count']:
            type_fill[3][0] = 'inter_auth'
        else:
            type_fill[3][0] = 'not_auth'

        if has_holder_type:
            for key in holder_config['holder_type_info'].keys():
                if key == 'hardware_type':
                    type_fill[0][0] = holder_config['holder_type_info']['hardware_type']
                elif key == 'hardware_subtype':
                    type_fill[0][1] = holder_config['holder_type_info']['hardware_subtype']
                elif key == 'hardware_brand':
                    type_fill[0][2] = holder_config['holder_type_info']['hardware_brand']
                elif key == 'os_category':
                    type_fill[1][0] = holder_config['holder_type_info']['os_category']
                elif key == 'os_subcategory':
                    type_fill[1][1] = holder_config['holder_type_info']['os_subcategory']
                elif key == 'os_dist':
                    type_fill[1][2] = holder_config['holder_type_info']['os_dist']
                elif key == 'os_subdist':
                    type_fill[1][3] = holder_config['holder_type_info']['os_subdist']
                elif key == 'account_type':
                    type_fill[2][0] = holder_config['holder_type_info']['account_type']
                elif key == 'account_subtype':
                    type_fill[2][1] = holder_config['holder_type_info']['account_subtype']
                elif key == 'ca_status':
                    type_fill[3][0] = holder_config['holder_type_info']['ca_status']

        type_fill = auto_fill_types(type_fill)
        assert type_fill is not None, (
            f'Manual configuration type settings for {holder_name} were not valid.\n'
            f'\t   Settings in question: {holder_config["holder_type_info"]}.\n'
            '\t   Please check that the type configuration of the holder does not'
            ' violate any rules laid out in CONFIG_GUIDE.md and alter the settings.'
        )

        self.holder_type_info: HOLDER_TYPE_INFO = HOLDER_TYPE_INFO(
            hardware_type=type_fill[0][0],
            hardware_subtype=type_fill[0][1],
            hardware_brand=type_fill[0][2],
            os_category=type_fill[1][0],
            os_subcategory=type_fill[1][1],
            os_dist=type_fill[1][2],
            os_subdist=type_fill[1][3],
            account_type=type_fill[2][0],
            account_subtype=type_fill[2][1],
            ca_status=type_fill[3][0]
        )

        # Holder information
        has_holder_info = 'holder_info' in holder_config.keys()
        if has_holder_info and 'common_name' in holder_config['holder_info'].keys():
            common_name = holder_config['holder_info']['common_name']
        else:
            common_name = 'device_' + hash_info(self.holder_type_info.long_name, self.env_info.uid_hash)[:16]

        if has_holder_info and 'country' in holder_config['holder_info'].keys():
            country = holder_config['holder_info']['country']
        else:
            country = get_random_country()

        if has_holder_info and 'state' in holder_config['holder_info'].keys():
            state = holder_config['holder_info']['state']
        else:
            state = 'State in ' + country

        if has_holder_info and 'locality' in holder_config['holder_info'].keys():
            locality = holder_config['holder_info']['locality']
        else:
            locality = 'Locality in ' + country

        if has_holder_info and 'org' in holder_config['holder_info'].keys():
            org = holder_config['holder_info']['org']
        else:
            org = common_name + "'s organization"

        if has_holder_info and 'org_unit' in holder_config['holder_info'].keys():
            org_unit = holder_config['holder_info']['org_unit']
        else:
            if self.holder_type_info.ca_status in ['inter_auth', 'root_auth']:
                org_unit = 'Certificates'
            else:
                org_unit = get_random_division()

        if has_holder_info and 'email' in holder_config['holder_info'].keys():
            email = holder_config['holder_info']['email']
            subdomain = email.split('@')[1]
        else:
            username = common_name.lower().replace(" ", "")
            subdomain = f'{org_unit.lower().replace(" ", "")}.theirorg.com'
            email = f'{username}@{subdomain}'

        if has_holder_info and 'url' in holder_config['holder_info'].keys():
            url = holder_config['holder_info']['url']
        else:
            if self.holder_type_info.ca_status == 'inter_auth':
                url = subdomain + '/intermediate_ca'
            elif self.holder_type_info.ca_status == 'root_auth':
                url = subdomain + '/root_ca'
            else:
                url = 'www.' + subdomain

        self.holder_info: HOLDER_INFO = HOLDER_INFO(
            common_name=common_name,
            country=country,
            state=state,
            local=locality,
            org=org,
            org_unit=org_unit,
            email=email,
            url=url
        )

        self.holder_info_hash: str = hash_info(self.holder_info.hash_content, self.env_info.uid_hash)

        # Create key pair
        self.holder_priv_key: Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey] = create_private_key(
            self.env_info.encrypt_alg
        )
        self.holder_pub_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey] = self.holder_priv_key.public_key()

        # Creating certificate variables
        self.holder_cert = None
        self.root_certs = {}
        self.cached_certs = {}

        # Create ports for receiving information
        self.csr_message_port: PriorityQueue = PriorityQueue()
        self.reg_message_port: PriorityQueue = PriorityQueue()
        self.ocsp_message_port: PriorityQueue = PriorityQueue()

        # Create flags
        self.has_root_cert_cache: bool = False
        self.has_self_cert: bool = False
        self.need_new_cert: bool = False
        self.cached_certs_empty: bool = False

        self.waiting_for_csr_response: bool = False
        self.waiting_to_send_csr: bool = False

        self.waiting_for_reg_response: bool = False
        self.waiting_to_send_reg: bool = False

        self.waiting_for_ocsp_response: bool = False
        self.waiting_to_send_oscp: bool = False

        # Create hub connection
        self.network_hub = None

        # Create CA-specific attributes
        if self.holder_type_info.ca_status in ['inter_auth', 'root_auth']:
            self.is_ca: bool = True
        else:
            self.is_ca: bool = False

        self.lower_level_certs: dict = {}
        self.cert_revoc_list: dict = {}

    def get_name(self) -> str:
        """
        Returns the name of the holder.

        Returns:
            str: Name of the holder.
        """

        return self.holder_name

    def set_hub_conn(self, hub) -> None:
        """
        Sets the connection to the hub.

        Args:
            hub: PKIHub - The hub to connect to.
        """

        self.network_hub = hub



"""
Module to test enumeration functionality.
"""

import unittest
import inspect
import random
import string

# Relative pathing from project root
import sys
from os.path import abspath, dirname, join

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

# Personal Modules must be imported after the system path is modified.
from PKIPractice.Utilities import EnumUtils


class TestEnums(unittest.TestCase):
    def test_enum_retrieval(self) -> None:
        """
        Tests the EnumUtils.get_all_items() function
        """
        def get_classes_from_module(module) -> list:
            """
            Retrieve the classes from the module as a list.
            """
            return [cls for name, cls in inspect.getmembers(module, inspect.isclass) if
                    cls.__module__ == module.__name__]

        enums = get_classes_from_module(EnumUtils)
        for enum in enums:
            self.assertIsNotNone(EnumUtils.get_all_items(enum))
            self.assertIsNotNone(EnumUtils.get_all_items(enum, True))

    def test_enum_default_values(self) -> None:
        """
        Tests that the values returned by EnumUtils.get_all_items() can be worked with.
        """
        def get_classes_from_module(module) -> list:
            """
            Retrieve the classes from the module as a list.
            """
            return [cls for name, cls in inspect.getmembers(module, inspect.isclass) if
                    cls.__module__ == module.__name__]

        enums = get_classes_from_module(EnumUtils)
        for enum in enums:
            info = EnumUtils.get_all_items(enum, True)
            for enum_name, enum_value in info.items():
                self.assertEqual(enum_value, enum[enum_name].value)

    def test_enum_value_type(self) -> None:
        """
        Tests that all values in EnumUtils are tuples of strings or strings.
        """
        def get_classes_from_module(module) -> list:
            """
            Retrieve the classes from the module as a list.
            """
            return [cls for name, cls in inspect.getmembers(module, inspect.isclass) if
                    cls.__module__ == module.__name__]

        enums = get_classes_from_module(EnumUtils)
        for enum in enums:
            info = EnumUtils.get_all_items(enum, True)
            for enum_name, enum_value in info.items():
                is_tuple_or_string = isinstance(enum_value, tuple) or isinstance(enum_value, str)
                self.assertTrue(is_tuple_or_string)

                if isinstance(enum_value, tuple):
                    for v in enum_value:
                        self.assertIsInstance(
                            v,
                            str,
                            f'{v} is not a string, it is a {type(v)}. This is in the {enum_name} enum for the '
                            f'{enum.__name__} class.'
                        )

    def test_fail_unknown_inputs(self) -> None:
        """
        Fuzzing tests that EnumUtils.has_value() returns False for non-existent enums.
        """
        def get_classes_from_module(module) -> list:
            """
            Retrieve the classes from the module as a list.
            """
            return [cls for name, cls in inspect.getmembers(module, inspect.isclass) if
                    cls.__module__ == module.__name__]

        def generate_random_string(min_length: int = 1, max_length: int = 20) -> str:
            """Generate a random string of random length."""
            length = random.randint(min_length, max_length)  # Random length between min_length and max_length
            return ''.join(random.choices(string.ascii_letters + string.digits + string.punctuation, k=length))

        enums = get_classes_from_module(EnumUtils)
        for enum in enums:
            for i in range(10):
                random_string = generate_random_string()
                self.assertFalse(EnumUtils.has_value(enum, random_string), f'Failed with {random_string}')

    def test_default_generation(self) -> None:
        """
        Random generation stress test to ensure nothing breaks.
        """

        # Creating test set
        test_set: list = [
            [['', '', ''], ['', '', '', ''], ['', ''], ['']],
            [['endpoint', '', ''], ['', '', '', ''], ['', ''], ['']],
            [['network', '', ''], ['', '', '', ''], ['', ''], ['']],
            [['appliance', '', ''], ['', '', '', ''], ['', ''], ['']],
            [['peripheral', '', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'desktop', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'laptop', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'server', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'phone', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'iot', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'switch', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'router', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'access_point', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'firewall', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'utm', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'usb_key', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'smart_card', ''], ['', '', '', ''], ['', ''], ['']],
            [['', 'external_storage', ''], ['', '', '', ''], ['', ''], ['']],
            [['', '', ''], ['microsoft', '', '', ''], ['', ''], ['']],
            [['', '', ''], ['unix', '', '', ''], ['', ''], ['']],
            [['', '', ''], ['mobile', '', '', ''], ['', ''], ['']],
            [['', '', ''], ['routing', '', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'windows', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'windows_server', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'linux', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'bsd', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'solaris', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'mac_os_x', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'android', '', ''], ['', ''], ['']],
            [['', '', ''], ['', 'ios', '', ''], ['', ''], ['']],
            [['', '', ''], ['', '', '', ''], ['user', ''], ['']],
            [['', '', ''], ['', '', '', ''], ['admin', ''], ['']],
            [['', '', ''], ['', '', '', ''], ['system', ''], ['']],
            [['', '', ''], ['', '', '', ''], ['', ''], ['not_auth']],
            [['', '', ''], ['', '', '', ''], ['', ''], ['inter_auth']],
            [['', '', ''], ['', '', '', ''], ['', ''], ['root_auth']],
        ]

        # Random stress test
        for test_fill in test_set:
            for i in range(400):
                self.assertIsNotNone(
                    EnumUtils.auto_fill_types(test_fill),
                    f'Test Fill that broke test: {test_fill}'
                )

    def test_random_generation(self) -> None:
        def generate_random_string(min_length: int = 1, max_length: int = 15) -> str:
            """Generate a random string of random length."""
            length = random.randint(min_length, max_length)  # Random length between min_length and max_length
            return ''.join(random.choices(string.ascii_letters + string.digits + string.punctuation, k=length))

        for i in range(1000):
            default_fill = [['', '', ''], ['', '', '', ''], ['', ''], ['']]

            # Pick a random list
            random_fill_index = random.randint(0, len(default_fill) - 1)
            random_fill = default_fill[random_fill_index]

            # Pick a random index from that
            random_index = random.randint(0, len(random_fill) - 1)

            # Set that point to a random value
            default_fill[random_fill_index][random_index] = generate_random_string()

            # Run assurance
            self.assertIsNotNone(
                EnumUtils.auto_fill_types(default_fill),
                f'Test Fill that broke test: {default_fill}'
            )


if __name__ == '__main__':
    unittest.main()
